!===============================================================================
! Copyright 2003-2016 Intel Corporation All Rights Reserved.
!
! The source code,  information  and material  ("Material") contained  herein is
! owned by Intel Corporation or its  suppliers or licensors,  and  title to such
! Material remains with Intel  Corporation or its  suppliers or  licensors.  The
! Material  contains  proprietary  information  of  Intel or  its suppliers  and
! licensors.  The Material is protected by  worldwide copyright  laws and treaty
! provisions.  No part  of  the  Material   may  be  used,  copied,  reproduced,
! modified, published,  uploaded, posted, transmitted,  distributed or disclosed
! in any way without Intel's prior express written permission.  No license under
! any patent,  copyright or other  intellectual property rights  in the Material
! is granted to  or  conferred  upon  you,  either   expressly,  by implication,
! inducement,  estoppel  or  otherwise.  Any  license   under such  intellectual
! property rights must be express and approved by Intel in writing.
!
! Unless otherwise agreed by Intel in writing,  you may not remove or alter this
! notice or  any  other  notice   embedded  in  Materials  by  Intel  or Intel's
! suppliers or licensors in any way.
!===============================================================================

!  Content:
!    Calculation of cross-product matrix Example Program Text
!******************************************************************************/

      include 'mkl_vsl.f90'
      include "errcheck.inc"
      include "generatedata.inc"
      include "statchars.inc"

      program MKL_VSL_TEST

      USE MKL_VSL_TYPE
      USE MKL_VSL

      integer,parameter :: DIM = 4        ! Task dimension
      integer,parameter :: NN  = 1000     ! Number of observations

      real(kind=4),parameter :: P_THRESHOLD = 0.01

      real(kind=4) C(DIM*DIM)
      data C / 1.0, 0.0, 0.0, 0.0,                                      &
     &         0.0, 1.0, 0.0, 0.0,                                      &
     &         0.0, 0.0, 1.0, 0.0,                                      &
     &         0.0, 0.0, 0.0, 1.0 /

      real(kind=4) a(DIM)
      data a / 5.0, 5.0, 5.0, 5.0 /

      TYPE(VSL_SS_TASK) task
      integer p
      integer n
      integer x_storage
      integer cov_storage
      integer cp_storage
      real(kind=4) x(DIM,NN)
      real(kind=4) mean(DIM)
      real(kind=4) cov(DIM,DIM), cp(DIM,DIM)
      integer i, j, order, errcode
      integer errnums, method
      integer(kind=8) estimate

      real(kind=4) pval_cov(DIM,DIM)

!     ***** Initializing parameters for Summary Statistics task *****
      p               = DIM
      n               = NN
      x_storage       = VSL_SS_MATRIX_STORAGE_COLS
      cov_storage     = VSL_SS_MATRIX_STORAGE_FULL
      cp_storage      = VSL_SS_MATRIX_STORAGE_FULL

!     ***** Generate data set using VSL GaussianMV RNG *****
      errcode = sGenerateGaussianMVData( p, n, x, a, C )
      call CheckVslError( errcode )

!     ***** Create Summary Statistics task *****
      errcode = vslsssnewtask( task, p, n, x_storage, x )
      call CheckVslError( errcode )


!    ***** Initialization of the task parameters related
!          to cross-product matrix computation *****/
      errcode = vslssseditcp(task, mean, cp=cp, cp_storage=cp_storage)
      call CheckVslError( errcode )

!    ***** Cross-product matrix is included in the list of estimates
!          to compute *****
      estimate = VSL_SS_CP
      method   = VSL_SS_METHOD_FAST

!     ***** Compute the estimates using FAST method *****
      errcode = vslssscompute( task, estimate, method )
      call CheckVslError( errcode )

!     ***** Edit task parameters for computation of covariance *****
      errcode = vslsssedittask( task,VSL_SS_ED_COV,cov )
      call CheckVslError( errcode )
      errcode = vslissedittask( task,VSL_SS_ED_COV_STORAGE,cov_storage)
      call CheckVslError( errcode )

!    ***** Convert cross-product matrix into correlation matrix *****
      estimate = VSL_SS_COV
      method = VSL_SS_METHOD_CP_TO_COVCOR
      errcode = vslssscompute( task, estimate, method );
      call CheckVslError( errcode )


!     ***** Testing stat characteristics of the computed estimates *****
!     ***** Compute p-values for covariance computed from cross-product *
      call sComputePvalsVariance( p, n, cov, C, pval_cov )
      call sComputePvalsCovariance( p, n, cov, C, pval_cov )

!     ***** Checking the validity of p-values for all estimates *****
      errnums = 0
      do i = 1, p
        do j = 1, i
          if (pval_cov(i, j) < P_THRESHOLD) then
            errnums = errnums + 1
          end if
        end do
      end do

!     ***** Printing results *****
      print 9, 'Task dimension :         ', p
      print 9, 'Number of observations : ', n
      print *, ''

!     ***** Printing computed cross-produt matrix *****
      print 10, '  Computed cross-product matrix               '
      print *, ''
      do i = 1, p
        write (*, 14), cp(i, 1:p)
        print *, ''
      end do
      print *, ''
      print *, ''


!     ***** Printing p-values for covariance matrix estimate *****
      print *, 'P-values for covariance obtained from cross-product'
      do i = 1, p
        write (*, 14), pval_cov(i, 1:p)
        print *, ''
      end do
      print *, ''

!     ***** Printing summary of the test *****
      if ( errnums == 0 ) then
        print *, ' All the computed estimates agree with theory'
      else
        print *, ' Error: At least one of the computed estimates',      &
     &           ' disagrees with theory'
        stop 1
      end if

!     ***** Delete Summary Statistics task *****
      errcode = vslssdeletetask( task )
      call CheckVslError( errcode )

      call MKL_FREE_BUFFERS()

9     format(A,I4)
10    format(A,$)
11    format(A,I1,A,$)
12    format(F11.6,A,$)
13    format(F11.6,A,F11.6,A,F11.6,A,$)
14    format(4F14.6,$)

      end
